# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Lists module."""

import contextlib
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QAbstractItemView, QFrame, QListWidget, QListWidgetItem  # pylint: disable=no-name-in-module


class ListWidgetBase(QListWidget):
    """Base list widget."""

    sig_item_selected = Signal(object)

    def __init__(self, *args, **kwargs):
        """Custom list widget environments."""
        super().__init__(*args, **kwargs)
        self._items = []
        self._current_item = None

        # Widget setup
        self.setAttribute(Qt.WA_MacShowFocusRect, False)
        self.setFrameStyle(QFrame.Plain)
        self.setFocusPolicy(Qt.NoFocus)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setMovement(QListWidget.Static)
        self.setResizeMode(QListWidget.Adjust)
        self.setSelectionMode(QAbstractItemView.NoSelection)
        self.setViewMode(QListWidget.ListMode)

    # --- Qt overrided methods
    # -------------------------------------------------------------------------
    def addItem(self, item):
        """Add an item to the list."""
        QListWidget.addItem(self, item)
        self._items.append(item)
        self.setItemWidget(item, item.widget)
        item.setSizeHint(item.widget_size())

        self.setup_item(item)

        if self._current_item is None:
            self._current_item = item
            self.setCurrentRow(0)
        else:
            item.set_selected(False)

    def insertItem(self, row, item):
        """Override Qt method."""
        QListWidget.insertItem(self, row, item)
        self._items = self._items[:row] + [item] + self._items[row:]
        self.setItemWidget(item, item.widget)
        item.setSizeHint(item.widget_size())
        self.setup_item(item)

        if self._current_item is None:
            self._current_item = item
            self.setCurrentRow(0)
        else:
            item.set_selected(False)

    def clear(self):
        """Clear all list items and references."""
        super().clear()
        self._items = []

    def count(self):
        """Override Qt Method."""
        return len(self._items)

    def setCurrentRow(self, row, loading=False):
        """Override Qt method."""
        for i, item in enumerate(self._items):
            if i == row:
                item.set_selected(True)
                item.set_loading(loading)
                self._current_item = item
            else:
                item.set_selected(False)
                item.set_loading(False)

    def currentItem(self):
        """Override Qt method."""
        return self._current_item

    def currentIndex(self):
        """Override Qt method."""
        if self._current_item in self._items:
            return self._items.index(self._current_item)
        return None

    def item(self, index):
        """Override Qt method."""
        return self._items[index]

    def items(self):
        """Override Qt method."""
        return self._items

    def scroll_to_item(self, item):
        """Scroll to item with focus."""
        try:
            self.scrollToItem(item)
        except RuntimeError:
            pass

    # --- Custom methods
    # -------------------------------------------------------------------------
    def setup_item(self, item):  # pylint: disable=unused-argument
        """Add additional logic after adding an item."""
        print(self, 'Must oveload this method')
        raise NotImplementedError

    def update_style_sheet(self):
        """Update custom CSS style sheet."""
        for item in self._items:
            with contextlib.suppress(BaseException):
                # This error is just in case the C++ object has been deleted and it is not crucial to log.
                item.update_style_sheet()
                item.setSizeHint(item.widget_size())

        self.update()
        self.repaint()


class ListWidgetItemBase(QListWidgetItem):
    """Base list widget item."""

    def set_loading(self, value):
        """Set loading status of widgets."""

    def set_selected(self, value):
        """Set selected status of widgets."""

    def update_style_sheet(self):
        """Set style sheet."""



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Base configuration management."""

import contextlib
import os
import shutil
import sys
import typing


# -----------------------------------------------------------------------------
# --- Configuration paths
# -----------------------------------------------------------------------------

CONFIG_NAME: typing.Final[str] = 'anaconda-navigator'
SUBFOLDER: typing.Final[str] = os.path.join('.anaconda', 'navigator')


def get_home_dir() -> str:
    """Return user home directory."""
    if os.name == 'nt':
        fallback: str = ''
        path_env_var: str
        path_env_vars: typing.Sequence[str] = ('HOME', 'APPDATA', 'USERPROFILE', 'TMP')

        # look through all available values
        for path_env_var in path_env_vars:
            current: str = os.environ.get(path_env_var, '')
            if (not current) or (not os.path.isdir(current)):
                continue

            # save first existing folder as a fallback value
            fallback = fallback or current

            # prefer options which already contain `.anaconda\navigator`
            inner: str = os.path.join(current, SUBFOLDER)
            if os.path.isdir(inner):
                return current

        # if there was any working option - use it
        if fallback:
            return fallback

    else:
        with contextlib.suppress(BaseException):
            return os.path.expanduser('~')

    raise RuntimeError('Please define environment variable $HOME')


def get_conf_path(filename: typing.Optional[str] = None) -> str:
    """Return absolute path for configuration file with specified filename."""
    result: str = os.path.join(get_home_dir(), SUBFOLDER)
    os.makedirs(result, exist_ok=True)

    if filename is not None:
        return os.path.join(result, filename)
    return result


def is_ubuntu() -> bool:
    """Detect if we are running in an Ubuntu-based distribution."""
    if sys.platform.startswith('linux'):
        try:
            stream: typing.TextIO
            with open('/etc/lsb-release', 'rt', encoding='utf-8') as stream:
                return 'ubuntu' in stream.read().lower()
        except OSError:
            pass
    return False


def is_gtk_desktop() -> bool:
    """Detect if we are running in a Gtk-based desktop."""
    if sys.platform.startswith('linux'):
        return any(map(os.environ.get('XDG_CURRENT_DESKTOP', '').startswith, ['Unity', 'GNOME', 'XFCE']))
    return False  # type: ignore


def fix_recursive_folder() -> None:
    """Check and fix `.anaconda/navigator/.anaconda/navigator` folder issue."""
    root: str = get_conf_path()
    file: str = f'{CONFIG_NAME}.ini'

    # check if expected configuration file already exists
    if os.path.isfile(os.path.join(root, file)):
        return

    # check if invalid configuration file already exists
    inner: str = os.path.join(root, SUBFOLDER)
    if not os.path.isfile(os.path.join(inner, file)):
        return

    # move old configurations to the new location
    name: str
    for name in os.listdir(inner):
        try:
            os.rename(os.path.join(inner, name), os.path.join(root, name))
        except OSError:
            pass

    # create a link, so old location shares the content with a new location
    try:
        inner = os.path.dirname(inner)
        root = os.path.dirname(root)
        shutil.rmtree(inner)
        os.symlink(root, inner)
    except OSError:
        pass



# -*- coding: utf-8 -*-

"""Core interfaces for Conda error solvers."""

from __future__ import annotations

__all__ = ['Solver', 'SimpleSolver', 'SolverCollection', 'POOL']

import abc
import collections.abc
import typing

from anaconda_navigator.utils import notifications
from anaconda_navigator.utils import solvers as common_solvers
from .. import types as conda_types


ErrorT_contra = typing.TypeVar('ErrorT_contra', bound='conda_types.CondaErrorOutput', contravariant=True)


class SolverFunction(typing.Protocol[ErrorT_contra]):  # pylint: disable=too-few-public-methods
    """General form of a function, that could be used as a solver."""

    def __call__(self, error: ErrorT_contra) -> notifications.NotificationSource:
        """
        Solve Conda error.

        :param error: Description of the error to solve.
        :return: Message about what was fixed.
        """


ErrorT = typing.TypeVar('ErrorT', bound='conda_types.CondaErrorOutput')


class Solver(typing.Generic[ErrorT], metaclass=abc.ABCMeta):  # pylint: disable=too-few-public-methods
    """
    Abstract base for all Conda error solvers.

    :param kwargs: Exact values to search for in the error body.
    """

    __slots__ = ('__arguments',)

    def __init__(self, **kwargs: typing.Any) -> None:
        """Initialize new :class:`~ErrorSolver` instance."""
        self.__arguments: typing.Final[collections.abc.Mapping[str, typing.Any]] = dict(kwargs)

    def solve(self, error: 'conda_types.CondaErrorOutput') -> notifications.NotificationSource:
        """
        Attempt solution of the `error`.

        :param error: Description of the error to solve.
        :return: Message about what was fixed. Might be :code:`None` if the solver is not applicable.
        """
        if self._applicable(error=error):
            return self._solve(error=typing.cast(ErrorT, error))
        return None

    def _applicable(self, error: 'conda_types.CondaErrorOutput') -> bool:
        """
        Check if this solver could be used to solve the `error`.

        This method might be overwritten to add custom checks.

        :param error: Description of the error to solve.
        :return: Check result.
        """
        return all(
            error.get(key, None) == value
            for key, value in self.__arguments.items()
        )

    @abc.abstractmethod
    def _solve(self, error: ErrorT) -> notifications.NotificationSource:
        """
        Solve Conda error.

        :param error: Description of the error to solve.
        :return: Message about what was fixed.
        """


class SimpleSolver(Solver[ErrorT], typing.Generic[ErrorT]):  # pylint: disable=too-few-public-methods
    """
    Custom :class:`~ErrorSolver`, which uses external function to solve an error.

    :param __function__: Function to use for error solving.
    :param kwargs: Exact values to search for in the error body.
    """

    __slots__ = ('__function',)

    def __init__(self, __function__: 'SolverFunction[ErrorT]', **kwargs: typing.Any) -> None:
        """Initialize new :class:`~SimpleErrorSolver` instance."""
        super().__init__(**kwargs)
        self.__function: typing.Final[SolverFunction[ErrorT]] = __function__

    def _solve(self, error: ErrorT) -> notifications.NotificationSource:
        """
        Solve Conda error.

        :param error: Description of the error to solve.
        :return: Message about what was fixed.
        """
        return self.__function(error=error)


class SolverCollection(common_solvers.SolverCollection[Solver[typing.Any]]):
    """Collection of solvers."""

    __slots__ = ()

    @typing.overload
    def register_function(
            self,
            function: 'SolverFunction[ErrorT]',
            *,
            filters: collections.abc.Mapping[str, typing.Any] | None = None,
            tags: str | collections.abc.Iterable[str] = (),
            unique_tags: str | collections.abc.Iterable[str] = (),
    ) -> SolverFunction[ErrorT]:
        """Register function as a solver."""

    @typing.overload
    def register_function(
            self,
            function: None = None,
            *,
            filters: collections.abc.Mapping[str, typing.Any] | None = None,
            tags: str | collections.abc.Iterable[str] = (),
            unique_tags: str | collections.abc.Iterable[str] = (),
    ) -> collections.abc.Callable[[SolverFunction[ErrorT]], SolverFunction[ErrorT]]:
        """Register function as a solver."""

    def register_function(self, function=None, filters=None, tags=(), unique_tags=()):
        """
        Register new solver.

        This method can be used as a decorator, as well as direct function. See :meth:`~SolverCollection.register` for
        more details.

        :param function: Function that should be registered.
        :param filters: Values that should be in the error body in order to apply current fix.
        :param tags: Optional collection of common tags.
        :param unique_tags: Optional collection of tags, that should be unique only for this particular solver.
        """
        if filters is None:
            filters = {}

        def wrapper(item: SolverFunction[ErrorT]) -> SolverFunction[ErrorT]:
            self.register(
                solver=SimpleSolver[ErrorT](item, **filters),
                tags=tags,
                unique_tags=unique_tags,
            )
            return item

        if function is None:
            return wrapper
        return wrapper(function)

    def solve(
            self,
            error: conda_types.CondaErrorOutput,
            *,
            tags: str | collections.abc.Iterable[str] | None = None,
    ) -> notifications.NotificationCollection:
        """
        Detect and solve issues for the `context`.

        :param error: Conda error to solve.
        :param tags: Limit to issue solvers with specific tags.
        :return: Iterator of details about solved issues.
        """
        result: notifications.NotificationCollection = notifications.NotificationCollection()

        solver_record: common_solvers.SolverRecord[Solver[typing.Any]]
        for solver_record in self.only(tags=tags):
            result.extend(
                notifications.NOTIFICATION_QUEUE.push(
                    message=solver_record.solver.solve(error),
                    caption='Broken Conda configuration',
                    tags='conda',
                ),
            )

        return result


POOL: typing.Final[SolverCollection] = SolverCollection()



# -*- coding: utf8 -*-

"""Integration with Heap analytics."""

from __future__ import annotations

__all__ = ['HeapProvider']

import itertools

import collections.abc
import typing

from . import core
from . import utilities

if typing.TYPE_CHECKING:
    import typing_extensions


T = typing.TypeVar('T')

Item: typing_extensions.TypeAlias = 'collections.abc.Mapping[str, typing.Any]'
Source: typing_extensions.TypeAlias = 'T | collections.abc.Iterable[T]'


class AccountProperties(typing.TypedDict, total=False):
    """Body for the :code:`add_account_properties` request."""

    account_id: typing_extensions.Required[str]
    properties: dict[str, str | int]


class Event(typing.TypedDict, total=False):
    """Body for the :code:`track` request."""

    identity: typing_extensions.Required[str]
    event: typing_extensions.Required[str]
    properties: dict[str, str | int]
    timestamp: str
    idempotency_key: str


class UserProperties(typing.TypedDict, total=False):
    """Body for the :code:`add_user_properties` request."""

    identity: typing_extensions.Required[str]
    properties: dict[str, str | int]


class ApiClient(utilities.ApiClient):
    """
    Client to the Heap analytics API.

    :param app_id: ID of the Heap Application to use for tracking.
    """

    __slots__ = ('__credentials',)

    def __init__(self, app_id: str, *, root_url: str = 'https://heapanalytics.com/api/') -> None:
        """Initialize new instance of a :class:`~ApiClient`."""
        super().__init__(root_url)

        self.__credentials: typing.Final[collections.abc.Mapping[str, str]] = {
            'app_id': str(app_id),
        }

    def add_account_properties(
            self,
            accounts: Source[AccountProperties],  # pylint: disable=invalid-sequence-index
    ) -> bool:
        """
        Attach custom account properties to users.

        More details: `Add Account Properties <https://developers.heap.io/reference/add-account-properties>`_.
        """
        return self.__collect('add_account_properties', 'accounts', accounts)

    def add_user_properties(self, users: Source[UserProperties]) -> bool:  # pylint: disable=invalid-sequence-index
        """
        Attach custom properties to any identified users from your servers.

        More details: `Add User Properties <https://developers.heap.io/reference/add-user-properties>`_.
        """
        return self.__collect('add_user_properties', 'users', users)

    def track(self, events: Source[Event]) -> bool:  # pylint: disable=invalid-sequence-index
        """
        Send custom events to Heap server-side.

        More details: `Add User Properties <https://developers.heap.io/reference/track-1>`_.
        """
        return self.__collect('track', 'events', events)

    def __collect(self, url: str, name: str, content: Source[Item]) -> bool:  # pylint: disable=invalid-sequence-index
        """Send request of a common format to the Heap."""
        query: dict[str, typing.Any] = {**self.__credentials}
        if isinstance(content, collections.abc.Mapping):
            query.update(content)
        else:
            query[name] = list(content)

        try:
            self._request('POST', url, json=query)
        except IOError:
            return False
        return True


class HeapProvider(core.Provider):
    """
    Custom provider to integrate with Heap Analytics.

    :param app_id: ID of the Heap Application to use for tracking.
    :param identity: Identity of the current user to track.
    :param alias: Custom alias of the provider.
    """

    __slots__ = ('__api',)

    def __init__(
            self,
            context: core.Context,
            app_id: str,
            *,
            alias: str = 'heap',
            root_url: str = 'https://heapanalytics.com/api/',
    ) -> None:
        """Initialize new instance of a :class:`~HeapProvider`."""
        super().__init__(alias=alias, context=context)

        self.__api: typing.Final[ApiClient] = ApiClient(app_id=app_id, root_url=root_url)

    def close(self) -> None:
        """Close provider."""

    def process(self, events: collections.abc.MutableSequence[core.Event]) -> None:
        """Process all pending events."""
        event: core.Event
        query_user_properties: dict[str, str | int]
        query_events: list[Event] | Event
        while events:
            query_user_properties = {}
            query_events = []
            for event in itertools.islice(events, 0, 1000):
                query_events.append({
                    'event': event.name,
                    'idempotency_key': event.key,
                    'identity': self.context.user_identity,
                    'properties': {**event.properties, 'session': self.context.session_identity},
                    'timestamp': event.timestamp.isoformat(),
                })
                query_user_properties.update(event.user_properties)

            if len(query_events) == 1:
                query_events = query_events[0]
            if not self.__api.track(query_events):
                return

            if query_user_properties:
                query_user: UserProperties = {
                    'identity': self.context.user_identity,
                    'properties': query_user_properties,
                }
                if not self.__api.add_user_properties(query_user):
                    events[:len(query_events)] = [
                        item
                        for item in events[:len(query_events)]
                        if item.user_properties
                    ]
                    return

            events[:len(query_events)] = []



# -*- coding: utf-8 -*-

"""Common interface for third-party applications."""

from __future__ import annotations

import typing
from .base import *
from .constants import *
from . import config_utils
from . import parsing_utils
if typing.TYPE_CHECKING:
    from anaconda_navigator.api import process
    from anaconda_navigator.config import user as user_config


class ApplicationCollection(typing.NamedTuple):
    """Collection of collected applications."""

    installable_apps: typing.Mapping[str, BaseInstallableApp]
    web_apps: typing.Mapping[str, BaseWebApp]
    app_patches: typing.Mapping[str, AppPatch]


APPLICATION_CACHE: typing.Optional[ApplicationCollection] = None


def get_applications(
        configuration: typing.Optional['user_config.UserConfig'] = None,
        process_api: typing.Optional['process.WorkerManager'] = None,
        *,
        cached: bool = False,
) -> ApplicationCollection:
    """Initialize all :class:`~BaseInstallableApp` instances for apps listed in root config."""
    global APPLICATION_CACHE  # pylint: disable=global-statement
    if cached and (APPLICATION_CACHE is not None):
        return APPLICATION_CACHE

    if configuration is None:
        raise TypeError('`configuration` must be provided')
    if process_api is None:
        raise TypeError('`process_api` must be provided')

    result: ApplicationCollection
    context = parsing_utils.ParsingContext(
        process_api=process_api,
        user_configuration=configuration,
    )
    apps: typing.Sequence[typing.Union[BaseApp, AppPatch]] = config_utils.load_configuration(context=context)
    APPLICATION_CACHE = result = ApplicationCollection(
        installable_apps={
            item.app_name: item
            for item in apps
            if isinstance(item, BaseInstallableApp)
        },
        web_apps={
            item.app_name: item
            for item in apps
            if isinstance(item, BaseWebApp)
        },
        app_patches={
            item.app_name: item
            for item in apps
            if isinstance(item, AppPatch)
        },
    )
    return result
